r""" This file is part of karel_robot package.

Karel World Tiles
==================

This file defines the pure class :class:`Tile` and its children:
  - :class:`Tile` through which Karel can pass freely
  - :class:`Wall` which blocks Karel
  - :class:`Beeper`, a tile with beepers for Karel to pick up

See the README in the package root folder for general take and the
one in this folder for details.

^^^^^^^
LICENSE
^^^^^^^

A GPLv3/later license applies::

    The karel_robot package is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the karel_robot package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""
from __future__ import annotations
from dataclasses import dataclass


@dataclass
class Tile:
    """ A tile in Karel's world map, by itself non-blocking, empty.

    Check for children classes with isinstance.
    """

    chars: str
    """ Character(s) representing the Tile in map. Also used when parsing. """
    blocking: bool
    """ Whether :class:`karel_robot.robot.Karel` can pass through the Tile. """

    def __str__(self):
        """ The character to be printed on screen.

        WARNING: overwrite if ``len(chars) > 1``.
        """
        return self.chars


@dataclass
class Empty(Tile):
    """ Wall to block Karel from passing. """

    chars: str = ".0"
    blocking: bool = False

    def __str__(self):
        return '.'

    def __repr__(self):
        return "Empty()"


@dataclass
class Wall(Tile):
    """ Wall to block Karel from passing. """

    chars: str = "#"
    blocking: bool = True

    def __repr__(self):
        return "Wall()"


@dataclass
class Treasure(Tile):
    """ Treasure chest that Karel can not pass through. """

    chars: str = "$"
    blocking: bool = True

    def __repr__(self):
        return "Treasure()"


@dataclass
class Beeper(Tile):
    """ A Tile from which Karel can `pick_beeper`. """

    count: int = 1
    chars: str = "123456789"  # 0 is not included <- b = 10k + 0
    blocking: bool = False

    def __str__(self):
        return str(self.count) if self.count <= 9 else "+"

    def __repr__(self):
        return f"Beeper({self.count})"



#!/usr/bin/env python3
""" Karel implementation in Python (with curses).

==============================================================================
                             Run Karel the Robot
==============================================================================

As a standalone app, this script opens a map and lets the user control Karel
using a keyboard. A custom map and other options may be specified.

Write your Karel program in Python by importing karel_robot.run and using the
simple functions like move() and turn_left(). Alternatively write recursive
(non Python) programs and run them with '--program example.ks'. See Examples.


==============================================================================
                                   Examples
==============================================================================

Following examples assume you are running karel from terminal and have Python
version at least 3.6 installed.

- **Command Karel using keyboard in a infinite empty world**::

    karel

- **Write a simple Python script commanding Karel**::

    from karel_robot.run import *
    # moves to the wall, places one beeper and stops
    while not front_is_blocked():
        move()
    put_beeper()

  Then run Karel in a 10x10 world like this::

    python3 programs/walk.py -x 10 -y 10

- **Write a recursive Karel program as a challenge**::

    DEFINE MAIN
        IFWALL PUT MOVE
        IFWALL SKIP MAIN
    END
    RUN MAIN

  Run Karel in a one line 200 squares long world::

    karel --program programs/easy/1_walk.ks -y 1 -x 200

  > For details see the karel_robot/parsers/interpret.py file.


------------------------------------------------------------------------------
                                     Note
------------------------------------------------------------------------------

The curses screen starts up upon import and closes on error or program end.


------------------------------------------------------------------------------
                              Exported functions
------------------------------------------------------------------------------

You can use simple functions in Python programs importing karel_robot.run:

======================== ===================================================
        Movement
======================== ===================================================
``move()``               Karel moves in the direction he is facing
``turn_left()``          Karel turns left
``turn_right()``         Karel turns right
======================== ===================================================

======================== ===================================================
        Beepers
======================== ===================================================
``pick_beeper()``        Karel tries to pick up a beeper
``put_beeper()``         Karel puts down a beeper (if he has any)
``beeper_is_present()``  True iff Karel stands on a beeper
======================== ===================================================

======================== ===================================================
         Walls
======================== ===================================================
``front_is_blocked()``   True iff Karel can't move forward
``front_is_treasure()``  True iff Karel stands in front of a Treasure
======================== ===================================================

======================== ===================================================
       Direction
======================== ===================================================
``facing_north()``       True iff Karel is facing north (``^``)
``facing_south()``       True iff Karel is facing south (``v``)
``facing_east()``        True iff Karel is facing east (``>``)
``facing_west()``        True iff Karel is facing west (``<``)
======================== ===================================================

========================= ==================================================
       Execution
========================= ==================================================
``set_beepers(None)``     Set Karel's beepers (``None`` is ∞)
``set_speed(100)``        Set how fast Karel moves, 0 to 100
``pause()``               Pause execution and wait for user
``message(text, pause)``  Show message to user and if ``pause`` wait.
``write_map(path)``       Writes the map (must be finite) to a file
``exit()``                End execution
========================= ==================================================

For details see the section of the file marked 'KAREL FUNCTIONS'.


==============================================================================
                                   LICENSE
==============================================================================

The karel_robot package is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

The karel_robot package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
the karel_robot package. If not, see `<https://www.gnu.org/licenses/>`_.
"""
from curses import endwin, KEY_UP, KEY_LEFT, KEY_RIGHT, KEY_HELP, KEY_RESIZE
from errno import EINVAL, EIO
from sys import stderr
from argparse import FileType
from typing import Optional
import os.path

from .. import *
from ..parsers import *


##############################################################################
#                               KAREL START-UP                               #
##############################################################################

parser = get_parser()

if __name__ in ["__main__", "karel_robot.run.main"]:
    if parser.usage:
        parser.usage += " [-p program.ks]"
    parser.add_argument(
        "-W",
        action="count",
        dest="wait",
        default=0,
        help="Waits for user keypress when stepping through --program. "
        "On two or more (-WW) shows the info, but do not wait for confirmation. "
        "On three or more (-WWW) writes only program message to --logfile. ",
    )
    parser.add_argument(
        "-p",
        "--program",
        type=FileType("r"),
        metavar="karel_program.ks",
        help="Parse and run text file with a (non Python) recursive Karel program.",
    )

_karel: Optional[Karel] = None
_karel_map: Optional[MapType] = None

window_opt: Optional[Window] = None
argv = parser.parse_args()

if argv.karelpos or argv.kareldir:
    _karel = Karel(position=argv.karelpos, facing=argv.kareldir or ">")

# Board is loaded and starts the curses window
try:
    if argv.karelmap is not None:
        _new_style = os.path.splitext(argv.karelmap.name)[1] == ".km2"
        _m = MapParser(lines=argv.karelmap, karel=_karel, new_style_map=argv.new_style_map or _new_style)
        _karel, _karel_map = _m.karel, _m.karel_map
        argv.karelmap.close()
        # overwrite default '-x' if not set or if '--ix' was set
        if not argv.x_map or argv.infinite_x:
            argv.x_map = None if argv.infinite_x else _m.width
        if not argv.y_map or argv.infinite_y:
            argv.y_map = None if argv.infinite_y else _m.height
    # Curse the window
    window_opt = Window(
        karel=_karel,
        tiles=_karel_map,
        x_map=argv.x_map,
        y_map=argv.y_map,
        lookahead=argv.lookahead,
        speed=argv.speed,
        output=argv.output,
    )
    if argv.beepers is not None:
        window_opt.karel.beepers = argv.beepers
except RobotError as e:
    print("Failed setting up the map:\n" + str(e), file=stderr)
    exit(EINVAL)
except BaseException as e:
    try:
        endwin()
    except BaseException:
        print("Exception unrelated to screen was raised:", file=stderr)
        raise e
    print("Probably a problem with curses window:\n" + str(e), file=stderr)
    exit(EIO)


if window_opt is None:
    print("Failed setting up the curses window", file=stderr)
    exit(EINVAL)  # Never happens, right?

window: Window = window_opt


##############################################################################
#                                STATUS LINE                                 #
##############################################################################


def status_line(command: str = "", force=False):
    """ Unsafe print/log status line. """
    if argv.verbose == 0 and not argv.logfile and not force:
        return

    karel = window.karel
    if not karel:
        print("Karel left the screen!", file=stderr)  # never happens, right?
        exit(EINVAL)

    text = (
        f"{command.ljust(5) if command else ''} "
        f"{(*karel.position,)} "
        f"{repr(window.karel_tile)} "
    )
    if argv.verbose == 2:
        text += f"View{(*window.offset,)} "
    if argv.logfile and (force or argv.wait != 3):
        print(text, file=argv.logfile)
    if force or argv.verbose and not (argv.program and argv.wait):
        window.message(text)


@screen(window, draw=True)
def toggle_status_line(status=None):
    """ Return current status line setting and set to next one or custom. """
    v = argv.verbose
    argv.verbose = (v + 1) % 3 if status is None else status
    status_line(f"TOGGLE{v}")
    return v


##############################################################################
#                              KAREL FUNCTIONS                               #
##############################################################################

# Movement
@screen(window, moved=True)
def move():
    """ Karel tries to move in the direction he is facing. """
    window.move()
    status_line("MOVE")


@screen(window)
def turn_left():
    """ Karel turns left. """
    window.karel.turn_left()
    status_line("LEFT")


@screen(window)
def turn_right():
    """ Karel turns right. """
    window.karel.turn_right()
    status_line("RIGHT")


# Beepers
@screen(window)
def pick_beeper():
    """ Karel tries to pick up a beeper. """
    window.pick_beeper()
    status_line("TAKE")


@screen(window)
def put_beeper():
    """ Karel puts down a beeper (if he has any). """
    window.put_beeper()
    status_line("PUT")


@screen(window, draw=None)
def beeper_is_present():
    """ True iff Karel stands on a beeper. """
    status_line("IF_MARK")
    return window.beeper_is_present()


# Walls
@screen(window, draw=None)
def front_is_blocked():
    """ True iff Karel can't move forward. """
    status_line("IF_WALL")
    return window.front_is_blocked()


@screen(window, draw=None)
def front_is_treasure():
    """ True iff Karel stands in front of a Treasure. """
    status_line("IF_GOLD")
    return window.front_is_treasure()


# KAREL_DIR
@screen(window, draw=None)
def facing_north():
    """ True iff Karel is facing north (^). """
    status_line("IF_NORTH")
    return window.karel.facing_north()


@screen(window, draw=None)
def facing_south():
    """ True iff Karel is facing south (v). """
    status_line("IF_SOUTH")
    return window.karel.facing_south()


@screen(window, draw=None)
def facing_east():
    """ True iff Karel is facing east (>). """
    status_line("IF_EAST")
    return window.karel.facing_east()


@screen(window, draw=None)
def facing_west():
    """ True iff Karel is facing west (<). """
    status_line("IF_WEST")
    return window.karel.facing_west()


# Execution
def set_speed(spd):
    """ Set the number of ticks per second. """
    window.speed = spd


def set_beepers(b=0):
    """ Set Karel's beepers, with None working as inf. """
    window.karel.beepers = None if b is None else max(0, int(b))


@screen(window, draw=None)
def pause():
    """ Pause execution, press any key to continue. """
    window.pause()


@screen(window, draw=None)
def screen_resize():
    """ Window readjusts self to include Karel on board. """
    window.screen_resize()
    status_line("RESIZE")


##############################################################################
#                                   OUTPUT                                   #
##############################################################################


@screen(window, draw=None)
def message(text="Press Q to quit, P to pause", paused=False, *, color=None):
    """ Show message to user in the message window (bottom line). """
    window.message(text, color=color, pause=paused)


@screen(window, draw=None)
def write_map(filepath: Optional[str] = None):
    """ Save map to file, if map is not infinite and user can write to file. """
    o = window.output
    if filepath:
        window.output = filepath
    window.save()
    window.output = o


##############################################################################
#                                   CHEATS                                   #
##############################################################################


def front_set_tile(tile):
    """ Set tile in front of Karel. """
    @screen(window, draw=True)
    def _front_set_tile():
        if not isinstance(tile, Tile):
            return ValueError(f"Can not set map Tile to {tile}")
        window.karel_facing = tile
        status_line(f"SET {tile}")
    return _front_set_tile


def karel_tile_beepers(n: int):
    """ Set the tile Karel is standing on to include exactly `n` beepers. """
    @screen(window)
    def _karel_tile_beepers():
        """ Set the tile Karel stands on to exactly `n` beepers. """
        if n < 1:
            window.karel_tile = Empty()
        else:
            window.karel_tile = Beeper(count=n)
        status_line(f"SET{n}")
    return _karel_tile_beepers


##############################################################################
#                              INTERACTIVE MODE                              #
##############################################################################

def interactive():
    """ Command Karel on a Board using your keyboard. """

    screen(window)(status_line)("INTERACTIVE", force=True)
    outer_speed = window.speed
    set_speed(None)  # Human lives are too short

    class _InteractiveStop(Exception):
        pass

    def _interactive_stop():
        raise _InteractiveStop

    handle = {
        # basic commands
        KEY_LEFT: KeyHandle(repeat=True, handle=turn_left),
        KEY_RIGHT: KeyHandle(repeat=True, handle=turn_right),
        KEY_UP: KeyHandle(repeat=True, handle=move),
        ord("q"): KeyHandle(repeat=False, handle=lambda: exit()),
        ord("u"): KeyHandle(repeat=True, handle=put_beeper),
        ord("i"): KeyHandle(repeat=True, handle=pick_beeper),
        # save
        ord("w"): KeyHandle(repeat=True, handle=write_map),
        # set tiles
        ord("#"): KeyHandle(repeat=True, handle=front_set_tile(Wall())),
        ord("."): KeyHandle(repeat=True, handle=front_set_tile(Empty())),
        ord("$"): KeyHandle(repeat=True, handle=front_set_tile(Treasure())),
        # stop interactive only
        ord("I"): KeyHandle(repeat=False, handle=_interactive_stop),
        # change verbosity
        ord("V"): KeyHandle(repeat=True, handle=toggle_status_line),
        # resize screen
        ord("R"): KeyHandle(repeat=True, handle=screen_resize),
        KEY_RESIZE: KeyHandle(repeat=True, handle=screen_resize),
        # user help
        KEY_HELP: KeyHandle(
            repeat=True,
            handle=lambda: message("Use arrows to move, i/u/q to pIck, pUt and Quit."),
        ),
    }
    for i in range(10):
        handle[ord('0') + i] = KeyHandle(repeat=True, handle=karel_tile_beepers(i))

    try:
        window.get_char(no_delay=False, handle=handle)
    except _InteractiveStop:
        pass
    set_speed(outer_speed)


window.handle[ord("I")] = KeyHandle(repeat=False, handle=interactive)


@screen(window, draw=None)
def run_program():
    """ Parse recursive Karel program and run it.
    """
    status_line("PARSING PROGRAM", force=True)

    def out(m):
        """ Tee to logfile if set. """
        if argv.wait:
            message(m, paused=argv.wait == 1)
        if argv.logfile:
            print(m, file=argv.logfile)

    p = Program(
        lines=argv.program,
        commands=Commands(
            skip=lambda: None,
            move=move,
            left=turn_left,
            right=turn_right,
            pick=pick_beeper,
            put=put_beeper,
        ),
        conditions=Conditions(
            ifwall=front_is_blocked,
            ifmark=beeper_is_present,
        ),
        confirm=out if argv.wait else None,
    )
    status_line("LOADED PROGRAM", force=True)
    p.run()
    argv.program.close()
    argv.program = None


def main():
    """ The function to be executed as a script. """
    if argv.program:
        run_program()
    interactive()


if __name__ == "__main__":
    main()



r""" This file is part of karel_robot package.

Karel Robot
===========

This package provides the classes for robot Karel and his map
of tiles, walls and beepers.

  - This does not import inner packages!

Importing this package does not start the screen and is meant
for testing and code reuse by anyone interested.

If you just want to write Karel programs, start with::

 from karel.run import *

^^^^^^^
LICENSE
^^^^^^^

A GPLv3/later license applies::

    The karel_robot package is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the karel_robot package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""
from .window import *

__all__ = [
    'Karel',
    'Tile',
    'Empty',
    'Wall',
    'Beeper',
    'Treasure',
    'MapType',
    'KarelMap',
    'Board',
    'BoardView',
    'Window',
    'WindowOpen',
    'screen',
    'RobotError',
    'Point',
    'KeyHandle',
    'KeysHandler',
]



""" Parse Karel map file.

Example simple map, with tiles and Karel represented by one character::

    ..>.1
    ....#

For more then 9 beepers or beepers under Karel, use header and spaces::

    KAREL 2 1 > 0
    .  .  .  .  .  #
    $  #  5  . 12  .

This will put Karel right on the ``5`` beepers and make him start with no
beepers in his bag (``0``). To make his beeper bag unlimited use ``N``.

==============================================================================
                                   LICENSE
==============================================================================

The karel_robot package is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

The karel_robot package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
the karel_robot package. If not, see `<https://www.gnu.org/licenses/>`_.
"""
from typing import Tuple, Dict, Iterable, Optional
from .. import *


class MapParser:
    """ Take lines/map file and make a robot and a map.

    Attributes:
        karel: the robot
        karel_map: mapping of (x,y) to tiles, leaving out Empty
        height: of the map
        width: of the map
    """
    def __init__(
        self,
        lines: Iterable[str],
        karel: Optional[Karel] = None,
        new_style_map: bool = False,
    ):
        """ Take lines/map file and make a robot and a map.

        Args:
            lines: of the map, e.g. open file or ['>.', '.1']
            karel: use this karel instead (better warning)
            new_style_map: space separated tiles and specified Karel and dimensions.
        Raises:
            RobotError: on incorrect map or if Karel is not present
        """
        self.karel: Karel = karel if karel else Karel()
        """ The robot, basically position * direction """
        self.karel_map: Dict[Tuple[int, int], Tile] = dict()
        """ Mapping of (x,y) to tiles, leaving out Empty """
        self.width: Optional[int] = None
        """ Width of the map, with None meaning no limit """
        self.height: Optional[int] = None
        """ Height of the map, with None meaning no limit """

        self._empty = Empty()
        self._wall = Wall()
        self._lines = iter(lines)
        self._new_style = new_style_map
        self._lineno = lambda y: y + self._new_style + 1
        if self._new_style:
            try:
                header = next(self._lines)
            except StopIteration:
                raise RobotError("line 0: Map empty!")
            if karel is None:
                self.karel = self.parse_new_map_header(header)
        self._parse()

    def _parse(self):
        y = None
        for y, line in enumerate(self._lines):
            self._parse_line(y, line)

        if y is None:
            raise RobotError("Map is empty!")
        self.height = y + 1
        if self.karel is None:
            raise RobotError("Karel must be placed on the map!")

    def _parse_line(self, y: int, line: str):
        """ Parse one line of map, splitting characters or on space if new-style."""
        tokens = line.split() if self._new_style else line.strip()

        if not tokens or self.width is not None and self.width != len(tokens):
            raise ValueError(f"line {self._lineno(y)}: "
                             "Karel map must be a rectangle!")
        self.width = len(tokens)

        x = None
        for x, char in enumerate(tokens):
            t = self._parse_tile(char, x, y)
            if t is not None and t is not self._empty:
                self.karel_map[x, y] = t
        if x is None:
            raise RobotError(f"line {self._lineno(y)}: Is Empty!")

    @staticmethod
    def parse_new_map_header(line: str) -> Karel:
        """ Parse the first line of map, that sets Karel.

        Raises:
            RobotError: on line not matching the form 'KAREL X Y > B', where
                        X,Y are coordinates, > is one of the directions
                        and B is number of beepers or 'N' for infinite
        """
        try:
            k, kx, ky, kd, kb = line.split()
            if k[0].lower() != "k":
                raise RobotError(f"Karel not set ({k} != KAREL)")
            if kd not in Karel.CHARS:
                raise ValueError()
            return Karel(
                position=Point(x=int(kx), y=int(ky)),
                facing=kd,
                beepers=int(kb) if kb != "N" else None,
            )
        except ValueError:
            raise RobotError("Map header not in form 'KAREL X Y > B'")

    def _parse_tile(self, token: str, x: int, y: int) -> Optional[Tile]:
        """ Parse one token, e.g. ``'.'`` or ``'#'`` or ``'1984'``.

        Raises:
            RobotError: on token that does not make valid Tile, e.g. '1Q94'.
        """
        if token in Karel.CHARS:
            if self.karel is None:
                self.karel = Karel(Point(x, y), token)
            elif self.karel is not None:
                raise RobotError(f"Karel is already on {self.karel.position}.")
            return
        for tile in self._empty, self._wall, Treasure():
            if token in tile.chars:
                return tile
        try:
            if int(token) > 0:
                return Beeper(count=int(token))
        except ValueError:
            pass
        raise RobotError(f"line {self._lineno(y)}: Invalid token '{token}' in column {x+1}.")



""" Parsers for Karel programs.

Currently there are these parsers:

  - :class:`MapParser`
  - ``get_parser`` for program options from user
  - :class:`Program` for parsing recursive non-Python Karel programs.

==============================================================================
                                   LICENSE
==============================================================================

The karel_robot package is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

The karel_robot package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
the karel_robot package. If not, see `<https://www.gnu.org/licenses/>`_.
"""
from .map import MapParser
from .cli_arguments import get_parser
from .interpret import Program, Commands, Conditions

__all__ = [
    'MapParser',
    'get_parser',
    'Program',
    'Commands',
    'Conditions',
]



r""" This file is part of karel_robot package.

Karel's Board
=============

This file defines the pure class :class:`KarelMap` and its children:
  - :class:`KarelMap` that stores and accesses the tiles
  - :class:`Board` focuses on tiles around Karel
  - :class:`BoardView` models a limited view on the map

See the README in the package root folder for general take and the
one in this folder for details.


LICENSE
^^^^^^^

A GPLv3/later license applies::

    The karel_robot package is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the karel_robot package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""
from __future__ import annotations
from typing import MutableMapping, Tuple
from .robot import *
from .tiles import *

MapType = MutableMapping[Tuple[int, int], Tile]
""" Type whose instance ``m`` support ``m[x,y]``. """


class KarelMap:
    """ Container for tiles on a square grid.

    Example map of Karel world::

      1..#...
      #....^.

    Legend:
      - multiple :class:`Empty` tiles (``.``)

      - one :class:`Beeper` (``1``) is on (0,0)

      - robot :class:`Karel` (``^``) is on (5,1) and on plain map
        his starting tile is :class:`Empty`

      - blocking :class:`Wall` (``#``) on (3,0) and on (0,1)

        - space outside of view is also considered to be :class:`Wall`

    A map is infinite by default and can also be infinite in one dimension only::

        >.1...11..111.1.1.11...11.11.1.11.1.11..1.11.1111.111.1.1

    Note that y-coordinate is flipped to ease parsing and printing to screen.
    """

    def __init__(
            self,
            x_map: Optional[int] = None,
            y_map: Optional[int] = None,
            tiles: Optional[MapType] = None,
    ):
        """ Setup by default infinite map of Karel's world.

        Args:
            x_map: The width of the map.
            y_map: The height of the map.
            tiles: Mutable map supporting tiles[x, y].
        """
        self.x_map: Optional[int] = x_map
        """ The x_view of the map. """
        self.y_map: Optional[int] = y_map
        """ The y_view of the map. """
        self._tiles: MapType = dict() if tiles is None else tiles
        """ Mutable map supporting ``_tiles[y][x]``. """

    def in_range(self, x: int, y: int):
        """ Check if map is unlimited or includes this coordinate. """
        def in_range(i, top):
            return top is None or 0 <= i < top
        return in_range(y, self.y_map) and in_range(x, self.x_map)

    def __getitem__(self, xy: Tuple[int, int]) -> Tile:
        """ Access tile at coordinate.

        If tiles have border (0, 0 to x_map, y_map) then accessing tiles
        beyond it returns :class:`Wall`. If tiles is a dictionary, then
        coordinates not contained in it returns :class:`Empty`.
        """
        if not self.in_range(*xy):
            return Wall()
        return self._tiles.get(xy, Empty())

    def __setitem__(self, xy: Tuple[int, int], tile: Tile) -> None:
        """ Set tile at coordinate. """
        if isinstance(tile, Empty) and xy in self._tiles:
            del self._tiles[xy]
        elif self.in_range(*xy):
            self._tiles[xy] = tile

    def __len__(self):
        """ Number of tiles. """
        if self.x_map is None or self.y_map is None:
            raise ValueError("This KarelMap is unbounded")
        return self.x_map * self.y_map

    def __bool__(self):
        """ Check if nonempty. """
        return self.x_map is None or self.y_map is None or len(self)


class Board(KarelMap):
    """ Manage :class:`Karel` on a board.
    """

    def __init__(
            self,
            karel: Optional[Karel] = None,
            x_map: Optional[int] = None,
            y_map: Optional[int] = None,
            tiles: Optional[MapType] = None,
    ):
        """ Setup by default infinite board with Karel the robot.

        Args:
            karel: The robot or default Karel if None.
            x_map: The width of the map.
            y_map: The height of the map.
            tiles: Mutable map supporting tiles[y][x], like list or dict
                   (if dict then indices not in tiles are considered Empty).

        Description copied from :class:`KarelMap`.
        """
        super().__init__(x_map, y_map, tiles)
        self.karel = karel or Karel()
        """ The robot living on the board. """
        if self[self.karel.position].blocking:
            raise RobotError("Karel can not move!")

    @property
    def karel_tile(self) -> Tile:
        """ The :class:`Tile` that :class:`Karel` is standing on. """
        return self[self.karel.position]

    @karel_tile.setter
    def karel_tile(self, tile: Tile):
        self[self.karel.position] = tile

    @property
    def karel_facing(self) -> Tile:
        """ The :class:`Tile` that :class:`Karel` is facing.
        Iff beyond board, then :class:`Wall`.
        """
        x = self.karel.position.x + self.karel.facing.x
        y = self.karel.position.y + self.karel.facing.y
        return self[x, y]

    @karel_facing.setter
    def karel_facing(self, tile: Tile):
        """ The :class:`Tile` that :class:`Karel` is facing.
        If beyond board, then ignored.
        """
        x = self.karel.position.x + self.karel.facing.x
        y = self.karel.position.y + self.karel.facing.y
        self[x, y] = tile

    def move(self):
        """Karel tries to move in the direction he is facing. """
        if self.front_is_blocked():
            raise RobotError("Can't move. Karel is blocked!")
        self.karel.move()
        return self

    def pick_beeper(self) -> Board:
        """ :class:`Karel` tries to pick up a :class:`Beeper`. """
        tile = self.karel_tile
        if isinstance(tile, Beeper):
            if tile.count > 1:
                tile.count -= 1
            else:
                self.karel_tile = Empty()
            self.karel.pick_beeper()
        else:
            raise RobotError(f"Can't pick Beeper from {repr(self.karel_tile)}")
        return self

    def put_beeper(self) -> Board:
        """ :class:`Karel` puts down a :class:`Beeper` (if he has any). """
        self.karel.put_beeper()
        tile = self.karel_tile
        if isinstance(tile, Beeper):
            tile.count += 1
        else:
            self.karel_tile = Beeper()
        return self

    def front_is_blocked(self) -> bool:
        """ True iff :class:`Karel` can't move forward. """
        return self.karel_facing.blocking

    def front_is_treasure(self) -> bool:
        """ True iff :class:`Karel` stands in front of a :class:`Treasure`. """
        return isinstance(self.karel_facing, Treasure)

    def beeper_is_present(self) -> bool:
        """ True iff :class:`Karel` stands on a :class:`Beeper`. """
        return isinstance(self.karel_tile, Beeper)


class BoardView(Board):
    """ Board with limited view, that shifts with :class:`Karel`.
    """

    def __init__(
            self,
            x_view: int,
            y_view: int,
            lookahead: int = 1,
            karel=None,
            x_map=None,
            y_map=None,
            tiles=None,
    ):
        """ Setup by default infinite board viewed through limited
        window, that shifts with :class:`Karel`.

        Args:
            x_view: The x-axis size of the window.
            y_view: The y-axis size of the window.
            lookahead: The number of fields visible ahead of Karel.
            karel: The robot or default Karel if None.
            x_map: The width of the map.
            y_map: The height of the map.
            tiles: Mutable map supporting tiles[y][x], like list or dict
                   (if dict then indices not in tiles are considered Empty).

        Description copied from :class:`Board` and :class:`KarelMap`.
        """
        super().__init__(karel, x_map, y_map, tiles)
        self.x_view = x_view
        """ The x_view of the view on Karel's world. """
        self.y_view = y_view
        """ The y_view of the view on Karel's world. """
        self.advanced = False
        """ Use to check if last call to move advanced the view. """
        self._lookahead: int = max(0, lookahead)
        """ The number of fields visible ahead of Karel """
        self.offset: Point = Point(0, 0)
        """ Top left tile position in the real map coordinates. """
        self.reset_offset()

    def reset_offset(self):
        """ Recalculate the top-left corner. """

        def offset(pos_karel, abs_max, rel_max):
            """ Calculate offset for one dimension. """
            # if 0 <= pos_karel - pos < rel_max:
            #    return pos
            if abs_max is None:  # shift the border only
                m = min(rel_max - 1, self._lookahead)
                if pos_karel < m:
                    return pos_karel - m
                return pos_karel - rel_max - m
            return max(0, min(pos_karel, abs_max - rel_max))

        self.offset = Point(
            x=offset(self.karel.position.x, self.x_map, self.x_view),
            y=offset(self.karel.position.y, self.y_map, self.y_view),
        )

    @property
    def karel_pos(self):
        """ Karel's position in the shifted coordinates. """
        xk, yk = self.karel.position
        x, y = self.offset
        return Point(x=xk - x, y=yk - y)

    def border_visible(self, direction: Optional[KAREL_DIR] = None):
        """ Check if border is in view. """
        if direction is None:
            direction = self.karel.to_dir()
        if direction == "<":
            return self.x_map is not None and self.offset.x == 0
        if direction == ">":
            return self.x_map == self.offset.x + self.x_view
        if direction == "^":
            return self.y_map is not None and self.offset.y == 0
        if direction == "v":
            return self.y_map == self.offset.y + self.y_view

    def karel_lookahead(self):
        """ Check if there if karel can advance the view ahead. """
        x, y = self.karel_pos
        direction = self.karel.to_dir()
        if direction == "<":
            return x <= self._lookahead
        if direction == ">":
            return x >= self.x_view - self._lookahead - 1
        if direction == "^":
            return y <= self._lookahead
        if direction == "v":
            return y >= self.y_view - self._lookahead - 1

    def move(self) -> BoardView:
        """ Move view in along with :class:`Karel`. """
        self.advanced = not self.border_visible() and self.karel_lookahead()
        if self.advanced:
            self.offset = Point(
                x=self.offset.x + self.karel.facing.x,
                y=self.offset.y + self.karel.facing.y,
            )
        super(BoardView, self).move()
        return self

    def get_pos(self, x, y):
        """ Translate from relative (view) position to absolute (map).

        This also works for **one tile** beyond boundary.
        """
        if -1 <= x <= self.x_view and -1 <= y <= self.y_view:
            return Point(x=x + self.offset.x, y=y + self.offset.y)
        raise RuntimeError("BoardView accessed beyond the boundary. ")

    def get_view(self, x, y):
        """ Access tile at shifted (view) coordinate. """
        return self[self.get_pos(x, y)]

    def set_view(self, x, y, tile):
        """ Set tile at shifted (view) coordinate. """
        self[self.get_pos(x, y)] = tile



"""
TODO ask IV003 for permission
"""
from dataclasses import dataclass, fields
from typing import Callable, List, Dict, Union, NamedTuple, Iterable, Optional, TypeVar

Action = Callable[[], None]
""" Action that can be called when command is run. """

Check = Callable[[], bool]
""" Check that can be performed when conditional is run. """

UserConfirm = Callable[[str], None]
""" Function that shows message to user and waits for response. """

Procedure = str
""" Procedure identification - currently its name. """


class Conditional(NamedTuple):
    """ Ternary operator 'IfThenElse' in this order. """

    condition: str
    met: Procedure
    fail: Procedure


Statement = Union[Procedure, Conditional]
""" Statements are procedure/command names and conditional statements.

They compose the body of procedure definition.
"""


@dataclass
class Exec:
    """ Name of executed procedure and position in its list of statements. """

    procedure: str
    index: int


@dataclass
class Commands:
    """ Command names and associated action to be executed. """

    skip: Action
    move: Action
    left: Action
    right: Action
    pick: Action
    put: Action


T_Commands = TypeVar("T_Commands", bound=Commands)


@dataclass
class Conditions:
    """ Check names available to program with associated functions returning bool.

    Attributes:
        ifwall: run on e.g. 'IFWALL SKIP MOVE'
        ifmark: run on e.g. 'IFMARK PICK MAIN'
    """

    ifwall: Check
    ifmark: Check


T_Conditions = TypeVar("T_Conditions", bound=Conditions)


class Program:
    """ Parse and run Karel the Robot recursive non Python programs.

    Attributes:
        main: Name of the procedure that will be first run.
        stack: Explicit stack of procedures to be executed, see run function.
        procedures: Defined procedures and their bodies.
    """

    _tokens_expected = {"define": 2, "run": 2}

    def __init__(
        self,
        lines: Iterable[str],
        commands: T_Commands,
        conditions: T_Conditions,
        confirm: Optional[UserConfirm] = None,
    ):
        self.main: Optional[Procedure] = None
        """ Name of the procedure that will be first run. """

        # Try to avoid Python stack overflow
        self.stack: List[Exec] = []
        """ Explicit stack of procedures to be executed, see run function. """

        self.procedures: Dict[Procedure, List[Statement]] = {}
        """ Defined procedures and their bodies."""

        self._commands: T_Commands = commands
        """ Commands that can be parsed and run. """

        self._conditions: T_Conditions = conditions
        """ Check names available to program with functions returning bool. """

        for k in fields(self._conditions):
            self._tokens_expected[k.name] = 3

        self._confirm: Optional[UserConfirm] = confirm
        """ Optionally ask user to confirm next command. """

        self._proc_to_check: Dict[Procedure, List[int]] = {}
        """ Temporarily store procedure names for later check. """

        self._current: Optional[List[Statement]] = None
        """ The currently defined procedure (temporary). """

        for lineno, line in enumerate(lines, start=1):
            self._parse_line(lineno, line)

        if self._current is not None:
            raise RuntimeError("Last defined procedure not ENDed.")

        del self._current  # No need for this after parsing

        self._check_procedures()
        del self._proc_to_check

    def _add_to_check(self, lineno: int, *names: str) -> None:
        """ Check name(s) later, commands are skipped immediately.

        Args:
            lineno: line number of the checked names
            *names: to be checked later
        """
        for name in names:
            if hasattr(self._commands, name):
                continue
            self._proc_to_check.setdefault(name, []).append(lineno)

    def _parse_line(self, lineno: int, line: str) -> None:
        """ Parse single line, discarding comments, whitespace and letter case.

        Raises:
            RuntimeError: on incorrect line or program structure
        """
        # discard comments and tokenize
        tokens: List[str] = line.partition("#")[0].lower().split()
        if not tokens:  # empty line
            return

        if len(tokens) != Program._tokens_expected.get(tokens[0], 1):
            raise RuntimeError(f"line {lineno}: wrong number of tokens")

        if self._current is None:
            self._parse_decl(lineno, tokens)
        elif tokens[0] == "define":
            raise RuntimeError(
                f"line {lineno}: last procedure not ENDed before DEFINE."
            )
        else:
            self._parse_statement(lineno, tokens)

    def _parse_decl(self, lineno: int, declaration: List[str]) -> None:
        """ Parse lines like 'DEFINE MAIN' and 'RUN MAIN'.

        Raises:
            RuntimeError: on redefinition of procedure or main procedure,
                          also on unrecognized declaration
        """
        if declaration[0] == "define":
            if declaration[1] in self.procedures:
                raise RuntimeError(
                    f"line {lineno}: multiple definitions of {declaration[1]}"
                )

            self._current = []
            self.procedures[declaration[1]] = self._current
        elif declaration[0] == "run":
            if self.main:
                raise RuntimeError(f"line {lineno}: multiple RUN declarations")
            self._add_to_check(lineno, declaration[1])
            self.main = declaration[1]
        else:
            raise RuntimeError(f"line {lineno}: unknown declaration {declaration[0]}")

    def _parse_statement(self, lineno: int, command: List[str]) -> None:
        """ Parse statements and check procedure name later.

        Statements accepted:
          - ``'MOVE'`` and other :class:`Commands`.
          - ``('IFWALL','PASS','MOVE')`` starting with :class:`Conditions`.
          - ``'END'`` of the defined procedure.
        """
        assert self._current is not None
        if command[0] == "end":
            if not self._current:
                raise RuntimeError(
                    f"line {lineno}: procedure cannot be empty, use SKIP."
                )
            self._current = None
        elif hasattr(self._conditions, command[0]):
            self._add_to_check(lineno, *command[1:])
            self._current.append(Conditional(*command))
        else:  # procedure call
            self._add_to_check(lineno, command[0])
            self._current.append(command[0])

    def _check_procedures(self) -> None:
        """ Check all the procedure names encountered.

        Raises:
            RuntimeError: on undefined procedure
        """
        for name, lines in self._proc_to_check.items():
            if name not in self.procedures:
                raise RuntimeError(
                    f"Procedure '{name}' not defined on lines "
                    f"{', '.join(map(str, lines))}"
                )

    def run(self) -> None:
        """ Start program from the procedure set with ``'RUN'``/``self.main``. """
        if self.main is None:
            raise RuntimeError("no RUN declaration")

        self.stack.append(Exec(self.main, 0))

        while self.stack:
            top = self.stack[-1]
            procedure_list = self.procedures[top.procedure]
            command = procedure_list[top.index]
            if self._confirm is not None:
                com_str = command if isinstance(command, str) else " ".join(command)
                self._confirm(
                    f"[{top.procedure.upper()}:{top.index},"
                    f"stack size:{len(self.stack)}] "
                    f"{com_str.upper()} "
                )
            top.index += 1  # advance program counter

            # optimisation: tail recursion (pop early)
            if top.index == len(procedure_list):
                self.stack.pop()

            self._exec(command)

    def _exec(self, command: Statement) -> None:
        """ Execute one statement, e.g. ``'MOVE'`` or ``'IFWALL PASS MAIN'``.

        Args:
            command: Either a procedure/command name or conditional statement
        """
        if isinstance(command, Conditional):
            self._exec(command[1 if getattr(self._conditions, command[0])() else 2])
        elif hasattr(self._commands, command):
            getattr(self._commands, command)()
        else:  # procedure call
            if command not in self.procedures:
                raise RuntimeError("procedure not found: " + command)

            self.stack.append(Exec(command, 0))



r""" This file is part of karel_robot package.

Karel the Robot
===============

This file defines the pure class :class:`Karel` to represent
a robot positioned on a square grid/tiled board.

See the README in the package root folder for general take and the
one in this folder for details.

^^^^^^^
LICENSE
^^^^^^^

A GPLv3/later license applies::

    The karel_robot package is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the karel_robot package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""
from __future__ import annotations
from typing import NamedTuple, Dict, Union, List, Optional

# wait for 3.8
KAREL_DIR = str  # Literal[">", "<", "^", "v"]


class Point(NamedTuple):
    """ Better way to access 2-vector of ints, otherwise same as tuple. """
    x: int
    y: int


class RobotError(RuntimeError):
    """Wrong execution of Karel program - e.g. hitting a Wall. """


class Karel:
    """ Karel the Robot.

    Karel is a shortsighted robot, that carries around
    and picks up beepers (possibly infinitely many) on
    a 2D grid.
    """

    DIRECTIONS: Dict[KAREL_DIR, Point] = {
        ">": Point(x=1, y=0),
        "^": Point(x=0, y=-1),
        "<": Point(x=-1, y=0),
        "v": Point(x=0, y=1),
    }

    CHARS: List[KAREL_DIR] = list(DIRECTIONS.keys())

    INV_DIR: Dict[Point, KAREL_DIR] = {v: k for k, v in DIRECTIONS.items()}

    def __init__(
        self,
        position: Union[Point, None] = None,
        facing: Union[Point, KAREL_DIR] = ">",
        beepers: int = None,
    ):
        self.position: Point = Point(0, 0) if position is None else Point(*position)
        """ Coordinates of tile that Karel is standing on. """

        if isinstance(facing, Point):
            self.facing: Point = facing
            """ Direction that Karel is facing. """
        elif facing in self.CHARS:
            self.facing = self.DIRECTIONS[facing]
        else:
            raise ValueError(f"Karel can not look in the '{facing}' direction")

        self.beepers: Optional[int] = beepers
        """ Number of beeper Karel can put down (if None then infinite). """

    def __repr__(self):
        return (
            f"Karel({self.position}, {self.to_dir()}"
            f"{'' if self.beepers is None else ', beepers=' + str(self.beepers)}"
            ")"
        )

    def to_dir(self) -> KAREL_DIR:
        """ Karel one character representation showing direction, e.g '>'. """
        return Karel.INV_DIR[self.facing]

    def move(self) -> Karel:
        """"Karel moves in the direction that he is facing. """
        self.position = Point(
            x=self.position.x + self.facing.x, y=self.position.y + self.facing.y
        )
        return self

    def turn_left(self) -> Karel:
        """ Karel turns 90° anti-clockwise. """
        self.facing = Point(self.facing.y, -self.facing.x)
        return self

    def turn_right(self) -> Karel:
        """ Karel turns 90° clockwise. """
        self.facing = Point(-self.facing.y, self.facing.x)
        return self

    def holding_beepers(self) -> bool:
        """ True if Karel has some or unlimited beepers. """
        return self.beepers is None or self.beepers > 0

    def facing_dir(self, direction: KAREL_DIR) -> bool:
        """ True if Karel is facing in the north direction. """
        return self.to_dir() == direction

    def pick_beeper(self) -> Karel:
        """ Increase beeper count if it is limited. """
        if self.beepers is not None:
            self.beepers += 1
        return self

    def put_beeper(self) -> Karel:
        """ Decrease beeper count if it is limited.

        :raises RobotError: on 0 beepers
        """
        if not self.holding_beepers():
            raise RobotError("Can't put beeper. Karel has none!")
        if self.beepers is not None:
            self.beepers -= 1
        return self



""" Karel interactive program visualization.

This package provides the following functions (__all__) for writing programs.
It is also used to interactively command Karel by keyboard.

==============================================================================
                                   LICENSE
==============================================================================

The karel_robot package is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

The karel_robot package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
the karel_robot package. If not, see `<https://www.gnu.org/licenses/>`_.
"""
from .main import *

__all__ = [
    "move",
    "turn_left",
    "turn_right",
    "pick_beeper",
    "put_beeper",
    "beeper_is_present",
    "facing_north",
    "facing_east",
    "facing_south",
    "facing_west",
    "front_is_blocked",
    "front_is_treasure",
    "set_speed",
    "set_beepers",
    "pause",
    "message",
    "interactive",
    "write_map",
    "toggle_status_line",
]



""" Karel the Robot (now in Python)

See the README in linked GitHub repository for more details.
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="karel_robot",
    version="1.0.0",
    author="Ondřej Šebek",
    author_email="xsebek@fi.muni.cz",
    description="Karel the Robot simple library and interactive executable",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xsebek/karel",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Environment :: Console :: Curses",
        "Topic :: Education",
        "Intended Audience :: Education ",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['karel=karel_robot.run.main:main']
    }
)



r""" This file is part of karel_robot package.

Window to Karel's world
=======================

This file defines the manager class Window, that draws the board
on screen and shows messages to the user on the bottom line.

See the README in the package root folder for general take and the
one in this folder for details.


LICENSE
^^^^^^^

A GPLv3/later license applies::

    The karel_robot package is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the karel_robot package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""
from __future__ import annotations
import io
from functools import wraps
from typing import Callable, Any
from time import sleep
import curses
from .board import *


class KeyHandle(NamedTuple):
    """ Response to user pressing key. """
    repeat: bool
    handle: Callable[[], Any]


KeysHandler = Dict[int, KeyHandle]
""" Dictionary for specifying response to user keypress. """


class Window(BoardView):
    """ Cursed window that draws a Board and interacts with user.

    Note that the curses screen is started on `__init__`.
    """

    # all colors are set in `start_screen`
    class Colors:
        """ Curses colors for printing in terminal.
        """

        wall: int
        empty: int
        karel: int
        beeper: int
        karel_beeper: int
        exception: int
        complete: int
        clear: int

    #################################################################

    def __init__(
        self,
        karel: Optional[Karel] = None,
        tiles: Optional[MapType] = None,
        x_map: Optional[int] = None,
        y_map: Optional[int] = None,
        speed: Optional[float] = 2.0,
        lookahead: int = 1,
        output: Optional[str] = None,
    ):
        """
        Args:
            karel (Karel): The robot or default Karel if None.
            x_map: The x_view of the map.
            y_map: The y_view of the map.
            tiles (Optional[Any]): Mutable map supporting tiles[y_map][x_map],
                   like list or dict (if dict then indices
                   not in tiles are considered Tile).
            speed: The number of ticks per second.
            lookahead: The number of fields visible ahead of Karel.

        Description copied from :class:`BoardView`, :class:`Board`
        and :class:`KarelMap`.
        """
        super().__init__(
            x_view=-1,  # set later
            y_view=-1,  # set later
            x_map=x_map,
            y_map=y_map,
            karel=karel,
            tiles=tiles,
            lookahead=lookahead,
        )
        self.speed: Optional[float] = max(0.0, speed) or None
        """ Roughly the number of ticks per second, ``None`` meaning no wait. """

        self.screen = curses.initscr()
        """ Cursed screen representing the whole terminal window. """
        self._del_screen = True

        self.y_screen, self.x_screen = self.screen.getmaxyx()
        """ Dimension of the whole terminal screen. """

        self.setup_screen()
        self._del_setup = True

        # reset the view dimensions, now that we know screen
        self.x_view, self.y_view = self.get_board_size()
        self.output: Optional[str] = output
        # NOTE: sub-windows share with parent the same positions, no overlay
        self.board_win = self.screen.subwin(self.y_view + 2, self.x_view + 2, 0, 0)
        """ The sub-window where Karel and his board is drawn. """

        self.message_win = curses.newwin(1, self.x_screen, self.y_screen - 1, 0)
        """ The sub-window where messages to user are printed. """
        self._last_message_len = 0

        self.handle: KeysHandler = {
            ord("q"): KeyHandle(repeat=False, handle=lambda: exit()),
            ord("p"): KeyHandle(repeat=False, handle=self.pause),
            curses.KEY_RESIZE: KeyHandle(repeat=True, handle=self.screen_resize),
        }
        """ What to do on user response, in ``key:(retry, function)`` format. """

        self.screen_resize()
        self.no_complete = False

    def draw(self):
        """ Draw the whole board. """
        self.board_win.border()
        for y in range(self.y_view):
            for x in range(self.x_view):
                self.draw_tile(x, y)
        self.draw_karel_tile().board_win.refresh()
        return self.get_char(no_delay=True, handle=self.handle)

    def redraw(self, moved=False):
        """ Redraw Karel's tile and the one he `moved` from."""
        self.draw_karel_tile(moved).board_win.refresh()
        return self.get_char(no_delay=True, handle=self.handle)

    def pause(self):
        """ Wait for keypress. """
        self.draw_exception("PAUSED")

    def move(self):
        """ Move Karel inside view and if he were to leave the screen,
            move it too.
        """
        if super(Window, self).move().advanced:
            self.draw()

    # IMPLEMENTATION DETAILS ########################################

    def setup_screen(self):
        """ Initialize cursed screen. """
        curses.noecho()  # turn off automatic echoing of keys to the screen
        curses.cbreak()  # react to keys instantly
        self.screen.keypad(True)  # curses process special keys
        curses.curs_set(False)  # make cursor invisible
        self.screen.timeout(1)
        self.setup_colors()

    def setup_colors(self):
        """ Setup curses colors and store them in :class:`Window.Colors`. """
        curses.start_color()
        # Window attribute : (Color foreground, optional Color background)
        colors = {
            "wall": (curses.COLOR_WHITE, curses.COLOR_WHITE),
            "empty": (curses.COLOR_YELLOW, None),
            "karel": (curses.COLOR_CYAN, None),
            "beeper": (curses.COLOR_GREEN, None),
            "karel_beeper": (curses.COLOR_RED, None),
            "exception": (curses.COLOR_RED, None),
            "complete": (curses.COLOR_GREEN, None),
            "clear": (curses.COLOR_BLACK, None),
        }
        for i, (attr, (col_fg, col_bg)) in enumerate(colors.items(), start=1):
            curses.init_pair(i, col_fg, col_bg or curses.COLOR_BLACK)
            setattr(self.Colors, attr, curses.color_pair(i))

    def screen_resize(self):
        """ Recalculate the dimensions and draw the board again. """
        sleep(1 / 20)
        self.board_win.clear()
        self.message_win.clear()
        self.y_screen, self.x_screen = self.screen.getmaxyx()
        self.x_view, self.y_view = self.get_board_size()
        self.board_win.resize(self.y_view + 2, self.x_view + 2)
        self.message_win.mvwin(self.y_screen - 1, 0)
        self.message_win.resize(1, self.x_screen)
        self.message_win.refresh()
        self.reset_offset()
        self.draw()

    def get_char(self, no_delay=True, restore=False, handle: KeysHandler = None):
        # TODO

        def handle_char(char):
            """ Run handling function and return if input should be read again. """
            handle[char].handle()
            return handle[char].repeat

        while True:
            self.screen.nodelay(no_delay)
            if no_delay:
                self.wait()
            ch = self.screen.getch()
            self.screen.nodelay(restore)
            if handle and ch in handle:
                if handle_char(ch):
                    continue
            elif handle and curses.KEY_HELP in handle:
                if handle_char(curses.KEY_HELP):
                    continue
            return ch

    def wait(self):
        """ If speed set, then wait for `1/speed` of a second. """
        if self.speed is not None:
            sleep(1 / self.speed)
        return self

    def get_board_size(self):
        # TODO
        """

        Note: self.*_screen must be set!

        :return: the size of view without border
        """
        y_view = self.y_screen - 3  # one line for messages
        x_view = self.x_screen - 2  # two for border
        if y_view < 1 or x_view < 1:
            raise RuntimeError(
                "Screen too small "
                f"{self.x_screen, self.y_screen}"
                f" for board {self.x_map, self.y_map} "
                "Minimum: 3columns 4rows"
            )
        if self.y_map:
            y_view = min(self.y_map, y_view)
        if self.x_map:
            x_view = min(self.x_map, x_view)
        return x_view, y_view

    # DRAWING #######################################################

    def draw_tile(self, x, y):
        """ Draw the tile on (x_map, y_map). """
        tile = self.get_view(x, y)
        if isinstance(tile, Wall):
            color = self.Colors.wall
        elif isinstance(tile, (Beeper, Treasure)):
            color = self.Colors.beeper
        else:
            color = self.Colors.empty
        self.board_win.addch(y + 1, x + 1, str(tile), color)
        return self

    def draw_karel_tile(self, moved=False):
        """ Draw the tile Karel is standing on and the one he ``moved`` from. """
        self.board_win.addch(
            self.karel_pos.y + 1,  # 1 border line
            self.karel_pos.x + 1,
            self.karel.to_dir(),
            self.Colors.karel_beeper if self.beeper_is_present() else self.Colors.karel,
        )
        if moved:
            x, y = self.karel_pos
            vx, vy = self.karel.facing
            self.draw_tile(x - vx, y - vy)
        return self

    def draw_exception(self, exception):
        """ Draw exception and wait for keypress.

        :raises SystemExit: exits the program if user presses the Q key
        """
        curses.beep()
        self.message(str(exception), self.Colors.exception, pause=True)
        self.message("")

    def message(self, text: str, color: int = None, pause: bool = False):
        """ Draw text to user. """
        if color is None:
            color = self.Colors.karel
        if pause:
            text = text + " Press any key to continue"
        self.message_win.insertln()
        if text:
            self.message_win.addstr(0, 0, text[: self.x_screen], color)
        self.message_win.refresh()
        if pause:
            handle = self.handle.copy()
            del handle[ord("p")]
            self.get_char(no_delay=False, handle=handle)

    def save(self):
        # TODO
        if not (self.output and self.x_map and self.y_map):
            return self.draw_exception(
                f"Can not save Map(X{self.x_map},Y{self.y_map})"
                f"to {self.output if self.output else 'non-specified output'}.",
            )

        def str2(t):
            """ Beeper(1) ~~> 1 and Empty ~~~> . """
            return str(t.count) if isinstance(t, Beeper) else str(t)

        try:
            with io.open(self.output, mode="w") as output:
                print(
                    f"KAREL {self.karel.position.x} {self.karel.position.y} "
                    f"{self.karel.to_dir()} "
                    f"{'N' if self.karel.beepers is None else self.karel.beepers} ",
                    file=output,
                )
                for y in range(self.y_map):
                    print(
                        " ".join(map(str2, (self[x, y] for x in range(self.x_map)))),
                        file=output,
                    )
        except IOError:
            self.draw_exception(f"Output to file {self.output} failed.")
        else:
            self.message(f"Saved the map to {self.output}.")

    def close_screen(self):
        """ DO NOT USE WINDOW AFTER THIS! """
        if hasattr(self, "no_complete") and not self.no_complete:
            self.no_complete = True
            self.message("Program Complete! Press any key to exit")
            self.get_char(no_delay=False)
        if hasattr(self, "_del_setup") and self._del_setup:
            # reverse curses terminal settings
            self._del_setup = False
            curses.nocbreak()
            self.screen.keypad(False)
            curses.echo()
        if hasattr(self, "_del_screen") and self._del_screen:
            # restore terminal
            self._del_screen = False
            curses.endwin()

    def __del__(self):
        """ Show complete text and close screen on program end.
        """
        self.close_screen()


class WindowOpen:
    """ Window manager to be used like ``open``.

    Usage::

        with WindowOpen(x_map=1, y_map=1) as w:
           w.move()
           # ...
    """

    def __init__(
        self,
        karel: Optional[Karel] = None,
        tiles: Optional[MapType] = None,
        x_map: Optional[int] = None,
        y_map: Optional[int] = None,
        speed: Optional[float] = 2.0,
        lookahead: int = 1,
        output: Optional[str] = None,
    ):
        """ Description copied from :class:`Window`.

        Args:
            karel (Karel): The robot or default Karel if None.
            x_map: The x_view of the map.
            y_map: The y_view of the map.
            tiles (Optional[Any]): Mutable map supporting tiles[y_map][x_map],
                   like list or dict (if dict then indices
                   not in tiles are considered Tile).
            speed: The number of ticks per second.
            lookahead: The number of fields visible ahead of Karel.
            output: The path to file where map can be saved.

        Returns:
            A :class:`Window` instance, that when exiting ``with`` block
            will optionally save the map to ``output`` and close screen.

        """
        self._w = Window(karel, tiles, x_map, y_map, speed, lookahead, output)

    def __enter__(self):
        return self._w

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._w.no_complete = exc_type is not None
        # curses error is BaseException, can't do anything with that
        if exc_type is None or issubclass(exc_type, Exception):
            if self._w.output is not None:
                self._w.save()
            self._w.draw_exception(exc_val)
        self._w.close_screen()
        return False


def screen(win: Window, moved: bool = False, draw: Optional[bool] = False):
    """ Safely execute function and redraw.

    Use this when the window should stay open upon success.

    Args:
        win: The window managing screen.
        moved: Whether Karel has moved (redraws last tile).
        draw: Draw whole screen (None means no drawing).
    Returns:
        Wrapper to safely execute function func.
    """
    if not isinstance(win, Window):
        raise RobotError("Supplied window is not a Window object, not initialized?")

    def dec_refresh(func):
        """ Inner decorator that only takes function and safely executes it. """
        @wraps(func)
        def wrap_refresh(*args, **kwargs):
            """ The actual wrapper for function working with Window. """
            try:
                result = func(*args, **kwargs)
                if draw is True:
                    win.draw()
                elif draw is False:
                    win.redraw(moved)
                return result
            except RobotError as ex:
                try:
                    win.draw_exception(ex)
                except BaseException as ex:
                    win.no_complete = True
                    win.close_screen()
                    raise ex
            except BaseException as ex:
                win.no_complete = True
                win.close_screen()
                raise ex

        return wrap_refresh

    return dec_refresh



# Copied from 'karel_run.py'
""" Karel implementation in Python (with curses).

As a standalone app, this script opens a map and lets the user control Karel
using a keyboard. A custom map and other options may be specified.

Write your Karel program in Python by importing karel_robot.run and using the
simple functions like move() and turn_left(). Alternatively write recursive
(non Python) program and run it with 'karel --program example.ks'.

LICENSE

The karel_robot package is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

The karel_robot package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
the karel_robot package. If not, see `<https://www.gnu.org/licenses/>`_.
"""
from argparse import ArgumentParser, FileType


def get_parser() -> ArgumentParser:
    """ The command line argument parser used by karel executable
    and scripts importing karel_robot.run.
    """
    karel_doc, karel_license = map(lambda x: x.strip(), __doc__.split("LICENSE"))
    parser = ArgumentParser(
        description=karel_doc,
        epilog=karel_license,
        usage="%(prog)s [--help] [OPTIONS] [-m karel_map.km]"
    )
    parser.add_argument(
        "--version",
        action="version",
        version="karel-robot 1.0",
        help="Show program's version number and exit."
    )
    parser.add_argument(
        "-x",
        "--cols",
        dest="x_map",
        type=int,
        default=None,
        metavar="X",
        help="The width of the real map (default infinite).",
    )
    parser.add_argument(
        "-y",
        "--rows",
        dest="y_map",
        type=int,
        default=None,
        metavar="Y",
        help="The height of the real map (default infinite).",
    )
    parser.add_argument(
        "--ix",
        "--infinite_x",
        dest="infinite_x",
        action="store_true",
        help="Force infinite map in the x-dimension."
    )
    parser.add_argument(
        "--iy",
        "--infinite_y",
        dest="infinite_y",
        action="store_true",
        help="Force infinite map in the y-dimension."
    )
    parser.add_argument(
        "-d",
        "--kareldir",
        metavar=">",
        help="The direction Karel starts with (one of > v < ^).",
    )
    parser.add_argument(
        "-k",
        "--karelpos",
        nargs=2,
        type=int,
        metavar="n",
        help="Sets the position Karel starts on.",
    )
    parser.add_argument(
        "-b",
        "--beepers",
        type=int,
        metavar="B",
        help="Sets the number of beepers Karel starts with.",
    )
    parser.add_argument(
        "-s",
        "--speed",
        type=float,
        default=3,
        metavar="S",
        help="Sets the number of ticks per second, 0 is no limit.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=0,
        dest="verbose",
        default=1,
        help="Hides the status line.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=2,
        default=1,
        help="Shows the status line and adds more info.",
    )
    parser.add_argument(
        "-l",
        "--logfile",
        type=FileType("a", bufsize=1),
        metavar="F.log",
        help="Writes logging information to the file (needs append permission).",
    )
    parser.add_argument(
        "-L",
        "--lookahead",
        default=1,
        type=int,
        metavar="L",
        help="Set the number of fields visible ahead of Karel.",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="out.km2",
        help="Set the output file, save it yourself (map must be finite).",
    )
    parser.add_argument(
        "-n", "--new_style_map",
        action="store_true",
        help="Map starts with 'KAREL X Y > B' and has spaces between tiles.",
    )
    parser.add_argument(
        "-m",
        "--karelmap",
        type=FileType("r"),
        metavar="karel_map.km",
        help="Text file with a map of karel world (overwrites -x, -y).",
    )
    return parser
